//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
// 

#ifndef __TTETHERNETMODELV2_DOUBLEBUFFER_H_
#define __TTETHERNETMODELV2_DOUBLEBUFFER_H_

#include <omnetpp.h>
#include <EtherFrame_m.h>
#include "CTBuffer.h"

namespace TTEthernetModel {

/**
 * @brief Base class for a double buffer class.
 *
 * Use the implementations BGDoubleBuffer, TTDoubleBuffer, RCDoubleBuffer.
 *
 * @sa BGDoubleBuffer, TTDoubleBuffer, RCDoubleBuffer, Buffer
 *
 * @ingroup Buffer
 */
class DoubleBuffer : public virtual Buffer
{
    public:
        /**
         * @brief Constructor
         */
        DoubleBuffer();
        /**
         * @brief Destructor
         */
        virtual ~DoubleBuffer();

    private:
        /**
         * The EtherFrame pointer is stored here
         */
        EtherFrame *frame;

    protected:
        /**
         * @brief Stores a new frame.
         *
         * The incoming EtherFrame pointer is stored. A previously stored frame is deleted
         * @param newFrame the new EtherFrame to store.
         */
        virtual void enqueue(EtherFrame *newFrame);

        /**
         * @brief Returns the frame that is stored in the buffer
         *
         * The stored frame is copied and the copy is being returned. If there is no EtherFrame
         * stored NULL is returned.
         *
         * @return Pointer to the copied frame or NULL if no frame is stored.
         */
        virtual EtherFrame* dequeue();
};
}
#endif
